#!/usr/bin/env python3
"""
Visualization Generation Module for Report 004 - Rhinovirus Serotypes

This module creates the serotype synchrony visualization from rhinovirus serotype data
generated by create_report_card_data.py.
"""

import sys
import os
from pathlib import Path
from typing import Dict, List, Any, Optional, Tuple
import pandas as pd
import numpy as np
import plotly.graph_objects as go
from datetime import datetime, timedelta
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn, BarColumn
import argparse
import gzip
import pickle

# Load environment variables from .env file if running locally
try:
    from dotenv import load_dotenv
    # Look for .env file in the project root
    env_path = Path(__file__).parent.parent.parent / '.env'
    if env_path.exists():
        load_dotenv(env_path)
except ImportError:
    # dotenv not available, rely on environment variables being set externally
    pass

# Import our data modules
from create_report_card_data import create_report_card_data
from retrieve_rhinovirus_data import retrieve_rhinovirus_data
from sites_config import get_public_sites, get_private_sites, get_display_name_for_location, get_location_identifier_to_name_mapping


# Environment Detection
def detect_environment() -> str:
    """Detect if running locally or in GitHub Actions."""
    return "github-actions" if os.getenv('GITHUB_ACTIONS') else "local"


def get_output_base_dir(environment: str) -> Path:
    """Get base output directory based on environment."""
    script_dir = Path(__file__).parent
    if environment == "github-actions":
        return script_dir / "github-actions-outputs"
    else:
        return script_dir / "local-outputs"


def filter_data_by_visibility(
    serotype_data: pd.DataFrame,
    visibility: str = 'private'
) -> pd.DataFrame:
    """
    Filter serotype data based on site visibility.
    
    Args:
        serotype_data: DataFrame containing serotype data
        visibility: 'public' or 'private'
        
    Returns:
        Filtered DataFrame
    """
    if visibility == 'public':
        allowed_sites = get_public_sites()
    else:
        allowed_sites = get_private_sites()
    
    # Filter data to only include allowed sites
    filtered_data = serotype_data[serotype_data['sewershed_location'].isin(allowed_sites)].copy()
    
    return filtered_data


def load_report_data(data_dir: Path) -> Dict[str, pd.DataFrame]:
    """Load report card data from saved files."""
    console = Console()
    data = {}
    
    # Try to load from pickle files first (faster)
    files_to_load = [
        ('serotype_data', 'serotype_data.pkl.gz'),
        ('serotype_sort_order', 'serotype_sort_order.pkl.gz'),
        ('surge_dates', 'surge_dates.pkl.gz'),
        ('summary_stats', 'summary_stats.pkl.gz')
    ]
    
    for data_name, filename in files_to_load:
        filepath = data_dir / filename
        if filepath.exists():
            try:
                with gzip.open(filepath, 'rb') as f:
                    data[data_name] = pickle.load(f)
                console.print(f"[green]✓ Loaded {data_name} from {filename}[/green]")
            except Exception as e:
                console.print(f"[yellow]Failed to load {filename}: {e}[/yellow]")
                # Try CSV as fallback
                csv_path = data_dir / filename.replace('.pkl.gz', '.csv')
                if csv_path.exists():
                    df = pd.read_csv(csv_path)
                    # Parse date columns if they exist
                    date_cols = ['sample_date', 'surge_date']
                    for col in date_cols:
                        if col in df.columns:
                            df[col] = pd.to_datetime(df[col])
                    data[data_name] = df
                    console.print(f"[green]✓ Loaded {data_name} from CSV[/green]")
    
    return data


def create_serotype_synchrony_plot(
    serotype_data: pd.DataFrame,
    surge_dates: pd.DataFrame,
    serotype_order: List[str],
    output_path: Path,
    rpb_threshold: float = 0.1,
    visibility: str = 'private'
) -> go.Figure:
    """
    Create a plot showing serotype synchrony across locations using a timeline approach.
    Each serotype is a row, and we show when it's active in different cities.
    """
    # Filter data by visibility first
    filtered_data = filter_data_by_visibility(serotype_data, visibility)
    
    # Use all serotypes from the sort order
    display_serotypes = serotype_order
    
    # Filter data to these serotypes
    plot_data = filtered_data[filtered_data['serotype'].isin(display_serotypes)].copy()
    
    # Create figure
    fig = go.Figure()
    
    # Define color palette for locations - use harmonious colors based on middle blue from heatmap
    locations = sorted(plot_data['sewershed_location'].unique())
    # Remove 'Air Samples' as it's not relevant
    locations = [loc for loc in locations if loc != 'Air Samples']
    
    # Base color: middle blue from heatmap color scale (#4682b4 - SteelBlue)
    # Create harmonious color palette around this blue
    color_palette = [
        '#4682b4',  # SteelBlue (base color for Columbia, MO)
        '#2e8b57',  # SeaGreen (analogous)
        '#cd853f',  # Peru (complementary warm)
        '#9370db',  # MediumPurple (triadic)
        '#ff6347',  # Tomato (warm contrast)
        '#20b2aa',  # LightSeaGreen (cool analogous)
        '#daa520',  # GoldenRod (warm secondary)
        '#8fbc8f',  # DarkSeaGreen (muted green)
        '#b8860b',  # DarkGoldenRod (earth tone)
        '#708090',  # SlateGray (neutral)
        '#d2691e',  # Chocolate (warm earth)
        '#4169e1',  # RoyalBlue (blue variation)
        '#da70d6',  # Orchid (purple variation)
    ]
    
    # Ensure Columbia, MO gets the base blue color
    location_colors = {}
    columbia_idx = None
    for i, loc in enumerate(locations):
        if 'Columbia' in loc and 'MO' in loc:
            columbia_idx = i
            break
    
    if columbia_idx is not None:
        # Put Columbia, MO first and assign base blue
        locations = [locations[columbia_idx]] + [loc for i, loc in enumerate(locations) if i != columbia_idx]
    
    location_colors = dict(zip(locations, color_palette[:len(locations)]))
    
    # Track which locations have been added to legend
    legend_shown = set()
    
    # For each serotype, plot activity periods
    for s_idx, serotype in enumerate(display_serotypes):
        serotype_data_filtered = plot_data[plot_data['serotype'] == serotype]
        
        for location in locations:
            loc_data = serotype_data_filtered[serotype_data_filtered['sewershed_location'] == location]
            
            if not loc_data.empty:
                # Find periods of activity (RPB > threshold)
                active_data = loc_data[loc_data['reads_per_billion'] > rpb_threshold].sort_values('sample_date')
                
                if not active_data.empty:
                    # Plot each active period
                    dates = active_data['sample_date'].values
                    rpb_values = active_data['reads_per_billion'].values
                    
                    # Create scatter plot with size based on RPB
                    # Use sqrt for size scaling to make differences more visible
                    # Reduced multiplier to prevent overlap between rows
                    sizes = np.sqrt(rpb_values) * 1.5  # Smaller circles
                    
                    # Show in legend if this is the first trace for this location
                    show_in_legend = location not in legend_shown
                    if show_in_legend:
                        legend_shown.add(location)
                    
                    fig.add_trace(go.Scatter(
                        x=dates,
                        y=[s_idx] * len(dates),
                        mode='markers',
                        marker=dict(
                            size=sizes,
                            color=location_colors[location],
                            opacity=0.5,  # Semi-transparent to show overlaps
                            line=dict(width=0.5, color='black')  # Thin border
                        ),
                        name=get_display_name_for_location(location),
                        legendgroup=location,
                        showlegend=show_in_legend,
                        visible="legendonly" if location != "Columbia, MO" else True,  # Columbia visible, others in legend only
                        hovertemplate=f'{serotype}<br>{get_display_name_for_location(location)}<br>Date: %{{x}}<br>RPB: %{{customdata:.1f}}<extra></extra>',
                        customdata=rpb_values
                    ))
    
    # Add empty traces for any locations not yet in legend (ensures all cities appear)
    for location in locations:
        if location not in legend_shown:
            fig.add_trace(go.Scatter(
                x=[],
                y=[],
                mode='markers',
                marker=dict(
                    size=10,
                    color=location_colors[location],
                    opacity=0.5,
                    line=dict(width=0.5, color='black')
                ),
                name=get_display_name_for_location(location),
                legendgroup=location,
                showlegend=True,
                visible="legendonly" if location != "Columbia, MO" else True,  # Columbia visible, others in legend only
                hoverinfo='skip'
            ))
    
    # Update layout (no title - explanation will be in credits.md)
    fig.update_layout(
        xaxis=dict(
            title="Date",
            tickangle=-45,
            rangeslider=dict(visible=True, thickness=0.05),
            range=[plot_data['sample_date'].min() - pd.Timedelta(days=7),
                   plot_data['sample_date'].max() + pd.Timedelta(days=7)] if not plot_data.empty else None
        ),
        yaxis=dict(
            title="Serotype (ordered by surge date)",
            tickmode='array',
            tickvals=list(range(len(display_serotypes))),
            ticktext=display_serotypes,
            autorange='reversed'  # Reverse the order as requested
        ),
        height=max(800, 50 + len(display_serotypes) * 25),  # Dynamic height based on number of serotypes
        width=1200,
        hovermode='closest',
        legend=dict(
            title=dict(
                text="Locations<br><sub>Click: toggle on/off<br>Double-click: isolate<br>→ then click to show all</sub>",
                font=dict(size=12)
            ),
            orientation="v",
            yanchor="top",
            y=0.98,
            xanchor="left",
            x=1.02,
            itemclick="toggle",  # Allow clicking to toggle visibility
            itemdoubleclick="toggleothers",  # Double-click to isolate
            bgcolor="rgba(255, 255, 255, 0.8)",  # Semi-transparent white background
            bordercolor="black",
            borderwidth=1
        ),
        margin=dict(l=150, r=200, t=100, b=100)
    )
    
    # Add vertical lines for months
    if not plot_data.empty and 'sample_date' in plot_data.columns:
        min_date = plot_data['sample_date'].min()
        max_date = plot_data['sample_date'].max()
        # Only add lines if dates are reasonable (after 2020)
        if min_date.year > 2020:
            for date in pd.date_range(min_date, max_date, freq='MS'):
                fig.add_vline(x=date, line_width=1, line_dash="dash", line_color="gray", opacity=0.3)
    
    # Save
    fig.write_html(output_path)
    
    return fig


def create_all_visualizations(
    report_data: Dict[str, pd.DataFrame],
    output_dir: Path,
    visibility: str = 'private'
) -> None:
    """
    Create the serotype synchrony visualization.
    
    Args:
        report_data: Dictionary containing all report data
        output_dir: Directory to save outputs
        visibility: 'public' or 'private' - determines which sites to include
    """
    console = Console()
    
    # Extract data
    serotype_data = report_data.get('serotype_data')
    serotype_sort_df = report_data.get('serotype_sort_order')
    surge_dates = report_data.get('surge_dates')
    
    if serotype_data is None or serotype_sort_df is None:
        console.print("[red]Missing required data for visualization[/red]")
        return
    
    # Get serotype order
    serotype_order = serotype_sort_df.sort_values('sort_index')['serotype'].tolist()
    
    # Create output directory
    output_dir.mkdir(exist_ok=True, parents=True)
    
    console.print(f"[cyan]Creating serotype synchrony visualization ({visibility} sites)...[/cyan]")
    
    # Create filename based on visibility
    filename = f"serotype_synchrony_{visibility}.html"
    
    # Create serotype synchrony plot
    create_serotype_synchrony_plot(
        serotype_data,
        surge_dates,
        serotype_order,
        output_dir / filename,
        visibility=visibility
    )
    console.print(f"[green]✓ Created serotype synchrony plot ({visibility}): {filename}[/green]")


def main():
    """Main entry point for visualization generation."""
    console = Console()
    
    parser = argparse.ArgumentParser(description="Generate rhinovirus serotype synchrony visualization")
    parser.add_argument('--data-dir', type=str, default='local-outputs',
                        help='Directory containing the report data')
    parser.add_argument('--output-dir', type=str, default=None,
                        help='Directory for visualization outputs')
    parser.add_argument('--visibility', type=str, choices=['public', 'private', 'both'], default='both',
                        help='Generate visualizations for public sites, private sites, or both')
    
    args = parser.parse_args()
    
    # Detect environment
    environment = detect_environment()
    console.print(f"[cyan]Running in {environment} environment[/cyan]")
    
    # Load report data
    console.print("[blue]Loading report data...[/blue]")
    data_dir = Path(args.data_dir)
    report_data = load_report_data(data_dir)
    
    if not report_data:
        console.print("[red]No report data found. Run create_report_card_data.py first.[/red]")
        return 1
    
    # Determine which visualizations to create
    visibilities_to_create = []
    if args.visibility == 'both':
        visibilities_to_create = ['private', 'public']
    else:
        visibilities_to_create = [args.visibility]
    
    # Create visualizations for each visibility type
    for visibility in visibilities_to_create:
        console.print(f"\n[blue]Creating {visibility} visualization...[/blue]")
        
        # Set output directory for this visibility
        if args.output_dir:
            base_output_dir = Path(args.output_dir)
        else:
            base_output_dir = get_output_base_dir(environment)
        
        if environment == "github-actions":
            # In GitHub Actions, create separate directories for public and private
            if visibility == 'public':
                output_dir = base_output_dir / "visualizations_public"
            else:
                output_dir = base_output_dir / "visualizations_private"
        else:
            # For local development, put everything in visualizations folder but with different filenames
            output_dir = base_output_dir / "visualizations"
        
        # Create visualization
        create_all_visualizations(report_data, output_dir, visibility)
    
    console.print(f"\n[bold green]✓ All visualizations complete![/bold green]")
    
    return 0


if __name__ == "__main__":
    exit_code = main()
    sys.exit(exit_code)